# Copyright (c) 2024, InfinityQ Technology, Inc.
import enum

class Precision(str, enum.Enum):
    """
    All precision level supported by TitanQ.

    Use ``Precision.AUTO`` when you don't know which one to use.
    """

    AUTO = 'auto'
    STANDARD = 'standard'
    HIGH = 'high'

    def __str__(self) -> str:
        return str(self.value)



# Copyright (c) 2024, InfinityQ Technology, Inc.

import abc
import enum
from typing import List, Optional, Tuple, Union
import numpy as np
from numpy._typing import NDArray

class Vtype(str, enum.Enum):
    """
    All variable types currently supported by the solver.

    ℹ️ **NOTE:** Bipolar variables are not directly supported,
    but :class:`tools.BipolarToBinary` can be used as an alternative.
    """

    BINARY = 'binary'
    INTEGER = 'integer'
    CONTINUOUS = 'continuous'

    def __str__(self) -> str:
        return str(self.value)
    

class ConstraintType(str, enum.Enum):
    """
    All constraint types currently supported in expression.
    """
    EQUAL = 'eq'
    GREATER_EQUAL = 'ge'
    GREATER = 'gt'
    LESSER_EQUAL = 'le'
    LESSER = 'lt'

    def __str__(self) -> str:
        return str(self.value)


class ArithmeticMixin(abc.ABC):
    """abstract class that overrides the operators"""

    def __mul__(self, other):
        return mul_dispatcher(self, other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __matmul__(self, other):
        return matmul_dispatcher(self, other)

    def __rmatmul__(self, other):
        return self.__matmul__(other)

    def __add__(self, other):
        return add_dispatcher(self, other)

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        return self + other * -1
    
    def __rsub__(self, other):
        return (self - other) * -1

    def __pow__(self, other):
        if other == 2:
            return self * self
        raise ValueError(f"Only powers of 2 are supported.")
    
class ConstraintMixin(abc.ABC):
    """abstract class that overrides the operators of constraints"""

    def __eq__(self, other):
        return ConstraintExpression(self-other, ConstraintType.EQUAL)
    def __gt__(self, other):
        return ConstraintExpression(self-other, ConstraintType.GREATER)
    def __ge__(self, other):
        return ConstraintExpression(self-other, ConstraintType.GREATER_EQUAL)
    def __lt__(self, other):
        return ConstraintExpression(self-other, ConstraintType.LESSER)
    def __le__(self, other):
        return ConstraintExpression(self-other, ConstraintType.LESSER_EQUAL)
    

class VariableVector(ArithmeticMixin, abc.ABC):
    """
    Object That represent a vector of variable to be optimized.
    """

    # Disable NumPy's (ufuncs) to avoid unintended operations on this class with array
    __array_ufunc__ = None
    def __init__(self, name: str, size: int) -> None:
        if size < 1:
            raise ValueError("Variable vector size cannot be less than 1")

        self._name = name
        self._size = size
        self._iter = 0


    def __len__(self) -> int:
        return self._size


    def __getitem__(self, other) -> 'SubVariable':
        return SubVariable(self, other)
    
    def __iter__(self):
        self._iter = 0
        return self

    def __next__(self):
        if self._iter >= self._size:
            raise StopIteration

        value = self[self._iter]
        self._iter += 1
        return value


    def name(self) -> str:
        """
        :return: Name of this variable vector.
        """
        return self._name
    
    def _fastSum(self) -> 'Expression':
        newExpression = Expression([None] * self._size)
        for index, subVar in enumerate(self):
            newExpression.terms()[index] = Term(subVar, None, 1)
        return newExpression

    @abc.abstractmethod
    def vtype(self) -> Vtype:
        """
        :return: Type of variable in the vector.
        """


    @abc.abstractmethod
    def variable_types_as_list(self) -> str:
        """
        :return: Generate a string of 'b', 'i' or 'c' depending on the variable type
        """


    @abc.abstractmethod
    def variable_bounds(self) -> NDArray:
        """
        :return: The variable bounds associated to this variable vector
        """


class BinaryVariableVector(VariableVector):
    def vtype(self) -> Vtype:
        return Vtype.BINARY


    def variable_types_as_list(self) -> str:
        return "b" * len(self)


    def variable_bounds(self) -> NDArray:
        return np.tile(np.array([0,1], dtype=np.float32), (self._size, 1))


class IntegerVariableVector(VariableVector):
    def __init__(self, name: str, size: int, variable_bounds: List[Tuple[int, int]]) -> None:
        super().__init__(name, size)

        if len(variable_bounds) != len(self):
            raise ValueError("variable_bounds need to be the same length as variable size")

        self._variable_bounds = np.array(variable_bounds, dtype=np.float32)


    def vtype(self) -> Vtype:
        return Vtype.INTEGER


    def variable_types_as_list(self) -> str:
        return "i" * len(self)


    def variable_bounds(self) -> NDArray:
        return self._variable_bounds

class ContinuousVariableVector(VariableVector):
    def __init__(self, name: str, size: int, variable_bounds: List[Tuple[int, int]]) -> None:
        super().__init__(name, size)

        if len(variable_bounds) != len(self):
            raise ValueError("variable_bounds need to be the same length as variable size")

        self._variable_bounds = np.array(variable_bounds, dtype=np.float32)


    def vtype(self) -> Vtype:
        return Vtype.CONTINUOUS


    def variable_types_as_list(self) -> str:
        return "c" * len(self)


    def variable_bounds(self) -> NDArray:
        return self._variable_bounds


##############################
# Expression related classes #
##############################
class SubVariable(ArithmeticMixin, ConstraintMixin):
    """sub variable of a variable vector"""
    __array_ufunc__ = None
    def __init__(
        self,
        parent: VariableVector,
        index: int
    ) -> None:
        self._parent = parent
        self._index = index

    def parent(self) -> VariableVector:
        return self._parent

    def index(self) -> int:
        return self._index


class Term(ArithmeticMixin, ConstraintMixin):
    __array_ufunc__ = None
    def __init__(
        self,
        v1: SubVariable,
        v2: Optional[SubVariable],
        coeff: float
    ) -> None:
        self._v1 = v1
        self._v2 = v2
        self._coeff = coeff

    def v1(self) -> SubVariable:
        return self._v1

    def v2(self) -> Optional[SubVariable]:
        return self._v2

    def coeff(self) -> float:
        return self._coeff


class Expression(ArithmeticMixin, ConstraintMixin):
    __array_ufunc__ = None
    def __init__(
        self,
        terms: List[Union[Term, float, int]] = []
    ) -> None:
        self._terms = terms

    def terms(self) -> List[Union[Term, float, int]]:
        return self._terms


class VectorExpression(ArithmeticMixin):
    __array_ufunc__= None
    def __init__(self, expressions: List[Expression]):
        self._expressions = expressions

    def __len__(self) -> int:
        return len(self._expressions)

    def __getitem__(self, other) -> Expression:
        return self.expressions()[other]

    def expressions(self) -> List[Expression]:
        return self._expressions
    
    def _fastSum(self) -> 'Expression':
        nbrOfTerms = 0
        for expr in self.expressions():
            nbrOfTerms+= len(expr.terms())
        newExpression = Expression([None] * nbrOfTerms)

        currentIndex = 0
        for expr in self.expressions():
            terms = expr.terms()
            newExpression.terms()[currentIndex:currentIndex+len(terms)] = terms
            currentIndex += len(terms)
        return newExpression


def add_dispatcher(
    lhs: Union[Expression, SubVariable, Term, VariableVector, VectorExpression],
    rhs: Union[np.ndarray, Expression, SubVariable, Term, VariableVector, VectorExpression]
) -> Union[Expression, Term, VectorExpression]:
    """addition dispatch to the right operation"""
    if isinstance(lhs, SubVariable):
        if isinstance(rhs, SubVariable):
            return Expression([Term(lhs, None, 1), Term(rhs, None, 1)])
        if is_numeric(rhs) or isinstance(rhs, Term):
            return Expression([Term(lhs, None, 1), rhs])
        if isinstance(rhs, Expression):
            return Expression([Term(lhs, None, 1)] + rhs.terms())

    if isinstance(lhs, Term):
        if is_numeric(rhs) or isinstance(rhs, Term):
            return Expression([lhs, rhs])

    if isinstance(lhs, Expression):
        if isinstance(rhs, Expression):
            return Expression(lhs.terms() + rhs.terms())
        if is_numeric(rhs) or isinstance(rhs, Term):
            return Expression([rhs] + lhs.terms())

    if isinstance(lhs, (SubVariable, Term, Expression)):
        if isinstance(rhs, (VariableVector, VectorExpression, np.ndarray)):
            if isinstance(rhs, np.ndarray):
                raise_if_not_one_dim(rhs)
            return VectorExpression([lhs + rhs[i] for i in range(len(rhs))])

    if isinstance(lhs, (VariableVector, VectorExpression)):
        if isinstance(rhs, (VariableVector, VectorExpression, np.ndarray)):
            if isinstance(rhs, np.ndarray):
                raise_if_not_one_dim(rhs)
            if len(lhs) != len(rhs):
                raise ValueError(f"Cannot perform addition, length ({len(lhs)}) does not match length ({len(rhs)}).")
            return VectorExpression([lhs[i]+rhs[i] for i in range(len(rhs))])
        if is_numeric(rhs):
            return VectorExpression([lhs[i] + rhs for i in range(len(lhs))])

    return NotImplemented


def mul_dispatcher(
    lhs: Union[Expression, SubVariable, Term, VariableVector, VectorExpression],
    rhs: Union[np.ndarray, Expression, SubVariable, Term, VariableVector, VectorExpression]
) -> Union[Expression, Term, VectorExpression]:
    """multiply dispatch to the right operation"""
    if isinstance(lhs, SubVariable):
        if isinstance(rhs, SubVariable):
            return Term(lhs, rhs, 1)
        if isinstance(rhs, Term):
            if rhs.v2() is not None:
                raise ValueError(f"TitanQ only supports polynomials of maximum degree 2, the input expression has degree too hight")
            return Term(rhs.v1(), lhs, rhs.coeff())
        if is_numeric(rhs):
            return Term(lhs, None, rhs)

    if isinstance(lhs, Term):
        if isinstance(rhs, Term):
            if lhs.v2() is not None or rhs.v2() is not None:
                raise ValueError(f"TitanQ only supports polynomials of maximum degree 2, the input expression has degree too hight")
            return Term(lhs.v1(), rhs.v1(), lhs.coeff()*rhs.coeff())
        if is_numeric(rhs):
            return Term(lhs.v1(), lhs.v2(), lhs.coeff()*rhs)

    if isinstance(lhs, Expression):
        if isinstance(rhs, Expression):
            new_expr = Expression()
            for term in lhs.terms():
                new_expr += term * rhs
            return new_expr
        if is_numeric(rhs) or isinstance(rhs, (Term, SubVariable)):
            return Expression([term*rhs for term in lhs.terms()])

    if isinstance(lhs, (SubVariable, Term, Expression)):
        if isinstance(rhs, (VariableVector, VectorExpression, np.ndarray)):
            if isinstance(rhs, np.ndarray):
                raise_if_not_one_dim(rhs)
            return VectorExpression([lhs*rhs[i] for i in range(len(rhs))])

    if isinstance(lhs, VariableVector):
        if isinstance(rhs, (VariableVector, np.ndarray)):
            if isinstance(rhs, np.ndarray):
                raise_if_not_one_dim(rhs)
            if len(lhs) != len(rhs):
                raise ValueError(f"Cannot perform multiplication, length ({len(lhs)}) does not match length ({len(rhs)}).")
            return VectorExpression([Expression([lhs[i]*rhs[i]]) for i in range(len(lhs))])
        if is_numeric(rhs):
            return VectorExpression([Expression([lhs[i]*rhs]) for i in range(len(lhs))])

    if isinstance(lhs, VectorExpression):
        if isinstance(rhs, (VariableVector, np.ndarray, VectorExpression)):
            if isinstance(rhs, np.ndarray):
                raise_if_not_one_dim(rhs)
            if len(lhs) != len(rhs):
                raise ValueError(f"Cannot perform multiplication, length ({len(lhs)}) does not match length ({len(rhs)}).")
            return VectorExpression([lhs[i]*rhs[i] for i in range(len(lhs))])
        if is_numeric(rhs):
            return VectorExpression([lhs[i]*rhs for i in range(len(lhs))])

    return NotImplemented


def matmul_dispatcher(
    lhs: Union[np.ndarray, VariableVector, VectorExpression],
    rhs: Union[np.ndarray, VariableVector, VectorExpression]
) -> Expression:
    """dot product dispatch to the right operation"""
    if isinstance(lhs, (VariableVector, np.ndarray, VectorExpression)):
        if isinstance(rhs, (VariableVector, np.ndarray, VectorExpression)):
            if len(lhs) != len(rhs):
                raise ValueError(f"Cannot perform dot product, length ({len(lhs)}) does not match length ({len(rhs)}).")
            if (isinstance(lhs, np.ndarray) or isinstance(rhs, np.ndarray)) and rhs.ndim == 2:
                return VectorExpression([lhs@rhs.T[i] for i in range(len(rhs.T))])
            if isinstance(lhs, VectorExpression) or isinstance(rhs, VectorExpression):
                terms_list= []
                [terms_list.extend((lhs[i] * rhs[i]).terms()) for i in range(len(lhs))]
                return Expression(terms_list)
            return Expression([lhs[i] * rhs[i] for i in range(len(lhs))])

    return NotImplemented


def raise_if_not_one_dim(array: np.ndarray) -> None:
    """raise ValueError if the array is larger than one dimension"""
    if array.ndim != 1:
        raise ValueError(f"Expected a 1-dimensional array, but got an array with {array.ndim} dimensions.")


def is_numeric(value) -> bool:
    """return if the value is either a float, int, a numpy int or a numpy float"""
    return isinstance(value, (float, int, np.integer, np.floating))


class ConstraintExpression():
    __array_ufunc__ = None
    def __init__(
        self,
        expression: Expression,
        constraintType: ConstraintType
    ) -> None:
        self._expression = expression
        self._constaintType = constraintType

    def expression(self) -> Expression:
        return self._expression
    
    def constraintType(self) -> ConstraintType:
        return self._constaintType


# Copyright (c) 2024, InfinityQ Technology, Inc.

class TitanqError(Exception):
    """Base TitanQ error"""

class MissingTitanqApiKey(TitanqError):
    """TitanQ Api key is missing"""

class MissingVariableError(TitanqError):
    """Variable has not already been registered"""

class VariableAlreadyExist(TitanqError):
    """Variable with the same name already exist"""

class MissingObjectiveError(TitanqError):
    """Objective has not already been registered"""

class MaximumConstraintLimitError(TitanqError):
    """The number of constraints is bigger than the number of variables"""

class ConstraintSizeError(TitanqError):
    """The constraint size does not match"""

class ConstraintAlreadySetError(TitanqError):
    """A constraint has already been set"""

class ObjectiveAlreadySetError(TitanqError):
    """An objective has already been set"""

class OptimizeError(TitanqError):
    """Error occur during optimization"""

class ServerError(TitanqError):
    """Error returned by the server"""

class ConnectionError(TitanqError):
    """Error due to a connection issue with an external resource"""

class MPSFileError(TitanqError):
    """Error due to an issue with the MPS file provided"""



# MIT License

# Copyright (c) 2019 Julian Märte

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

import re
import math
import copy
import warnings

CORE_FILE_ROW_MODE = "ROWS"
CORE_FILE_COL_MODE = "COLUMNS"
CORE_FILE_RHS_MODE = "RHS"
CORE_FILE_BOUNDS_MODE = "BOUNDS"
CORE_FILE_RANGES_MODE = "RANGES"
CORE_FILE_SOS_MODE = "SOS"

# SMPS special tags
CORE_FILE_ARCS_MODE = "ARCS"

CORE_FILE_BOUNDS_MODE_NAME_GIVEN = "BOUNDS_NAME"
CORE_FILE_BOUNDS_MODE_NO_NAME = "BOUNDS_NO_NAME"
CORE_FILE_RHS_MODE_NAME_GIVEN = "RHS_NAME"
CORE_FILE_RHS_MODE_NO_NAME = "RHS_NO_NAME"
CORE_FILE_RANGES_MODE_NAME_GIVEN = "RANGES_NAME"
CORE_FILE_RANGES_MODE_NO_NAME = "RANGES_NO_NAME"
CORE_FILE_SOS_MODE_FIRST_LINE = "SOS_FIRST_LINE"


ROW_MODE_OBJ = "N"

MODE = {CORE_FILE_ROW_MODE: 0, CORE_FILE_COL_MODE: 1, CORE_FILE_RHS_MODE: 3,
         CORE_FILE_BOUNDS_MODE: 4, CORE_FILE_BOUNDS_MODE_NAME_GIVEN: 5,
         CORE_FILE_BOUNDS_MODE_NO_NAME: 6, CORE_FILE_RHS_MODE_NAME_GIVEN: 7,
         CORE_FILE_RHS_MODE_NO_NAME: 8, CORE_FILE_RANGES_MODE_NAME_GIVEN: 9,
         CORE_FILE_RANGES_MODE_NO_NAME: 10, CORE_FILE_SOS_MODE: 11,
         CORE_FILE_SOS_MODE_FIRST_LINE: 12}

COL_TYPES = {1: "Integer", 0: "Continuous"}
BND_TYPES = ["LO", "UP"]
BND_TYPE = {"LO": "lower", "UP": "upper"}
BND_FUNC = {"LO": lambda a, b : a if b == 0 else max(a,b), "UP": min}

TIME_FILE_PERIODS_MODE = "PERIODS"
TIME_FILE_PERIODS_MODE_EXPLICIT = "PERIODS_EXPLICIT"
TIME_FILE_PERIODS_MODE_IMPLICIT = "PERIODS_IMPLICIT"
TIME_FILE_ROWS_MODE = "ROWS"
TIME_FILE_COLS_MODE = "COLUMNS"

STOCH_FILE_SIMPLE_MODE = "SIMPLE"
STOCH_FILE_CHANCE_MODE = "CHANCE"
STOCH_FILE_SCENARIOS_MODE = "SCENARIOS"
STOCH_FILE_NODES_MODE = "NODES"
STOCH_FILE_INDEP_MODE = "INDEP"
STOCH_FILE_BLOCKS_MODE = "BLOCKS"
STOCH_FILE_DISTRIB_MODE = "DISTRIB"
STOCH_FILE_BLOCKS_BL_MODE = "BL"

# BLOCKS modes
STOCH_FILE_BLOCKS_DISCRETE_MODE = "BLOCKS_DISCRETE"
STOCH_FILE_BLOCKS_SUB_MODE = "BLOCKS_SUB"
STOCH_FILE_BLOCKS_LINTR_MODE = "BLOCKS_LINTR"

#INDEP modes
STOCH_FILE_INDEP_DISCRETE_MODE = "INDEP_DISCRETE"
STOCH_FILE_INDEP_DISTRIB_MODE = "INDEP_DISTRIB"


warnings.simplefilter("always", ImportWarning)
# warnings.warn("""pysmps: Backwards compatibility to pre-2.0 versions has been abolished. Install older versions for code adjusted to the old interface, e.g. via \r\n
#               pip install 'pysmps<2.0'\r\n""", category=ImportWarning, stacklevel=2)
warnings.simplefilter("default", ImportWarning)

# public
class MPS:
    
    def __init__(self, defaults):
        self.defaults = {"Integer": {"type": "Integer", "lower": defaults["i_lower"], "upper": defaults["i_upper"]}, "Continuous": {"type": "Continuous", "lower": defaults["c_lower"], "upper": defaults["c_upper"]}}
        self.name = "No Name"
        self.objectives = {}
        self.variables = {}
        self._variables = {}
        self.constraints = {}
        self.rhs = {}
        self.offsets = {}
        self.ranges = {}
        #self.sos = {}
        
        self.curr_rhs = -1
        self.curr_bnd = -1
        self.curr_range = -1
        
        
    def finalize(self):
        if len(self.bnd_names()) > 0: self.curr_bnd = self.bnd_names()[0]
        if len(self.rhs_names()) > 0: self.curr_rhs = self.rhs_names()[0]
        if len(self.range_names()) > 0: self.curr_range = self.range_names()[0]
        #if len(self.sos_names()) > 0: self.curr_sos = self.sos_names()[0]
        
    def _set_name(self, name):
        self.name = name
    
    def add_objective(self, name):
        self.objectives[name] = {"coefficients": {}}
    
        
    # returns the list of names of possible boundary configurations
    def objective_names(self):
        return list(self.objectives.keys())
    def bnd_names(self):
        return list(self.variables.keys())
    def rhs_names(self):
        return list(self.rhs.keys())
    def variable_names(self):
        return list(self._variables.keys())
    def constraint_names(self):
        return list(self.constraints.keys())
    def range_names(self):
        return list(self.ranges.keys())
    #def sos_names(self):
    #    return list(self.sos.keys())
    
    
    
    def get_variables(self):
        return self.variables[self.curr_bnd]
    def get_rhs(self):
        return self.rhs[self.curr_rhs]
    def get_constraints(self):
        return self.constraints
    def get_objectives(self):
        return self.objectives
    def get_offsets(self):
        return self.offsets[self.curr_rhs]
    def get_ranges(self):
        return self.ranges[self.curr_range]
    
    
    def _add_constraint(self, name, constr):
        self.constraints[name] = constr
        
    def _set_coefficient(self, row, variable, value):
        if row in self.objectives:
            self.objectives[row]["coefficients"][variable] = value
        else:
            self.constraints[row]["coefficients"][variable] = value
    
    
    #def add_sos(self, name, degree):
    #    if name in self.sos:
    #        raise ValueError('The SOS ' + name + ' does already exist!')
    #    self.sos[name] = {"degree": degree, "priority": priority, "variables": {}}
    
    #def attach_sos(self, name):
    #    self.curr_sos = name
        
    #def add_variable_to_sos(self, var, value):
    #    self.sos[self.curr_sos]["variables"][var] = value
    
    # sets the boundary group active for modification
    def attach_bnd(self, name):
        if name not in self.variables:
            raise ValueError('The boundary group ' + name + ' does not exist!')
        self.curr_bnd = name
    
    def _add_variable(self, name, _type):
        self._variables[name] = {"type": _type, "lower": self.defaults[_type]["lower"], "upper": self.defaults[_type]["upper"]}
    def _update_variable(self, variable, mod):
        self.variables[self.curr_bnd][variable].update(mod)
        
    def _add_bnd_group(self, name):
        if name in self.variables:
            raise ValueError('The boundary group ' + name + ' already exists!')
        self.variables[name] = copy.deepcopy(self._variables)
        if len(self.variables) == 2:
            warnings.warn("There are multiple boundary groups for the linear program " + self.name + ". Switch boundary group using the MPS.set_bounds_group function.", stacklevel=2)
    
    def attach_range(self, name):
        if name not in self.ranges:
            raise ValueError('The range group ' + name + ' does not exist!')
        self.curr_range = name
    
    def _add_range_group(self, name):
        if name in self.ranges:
            raise ValueError('The range group ' + name + ' already exists!')
        self.ranges[name] = {}
        if len(self.ranges) == 2:
            warnings.warn("There are multiple range groups for the linear program " + self.name + ". Switch range group using the MPS.attach_range function.", stacklevel=2)
    
    def add_range(self, constr, value):
        if self.constraints[constr]["type"] == "G":
            self.ranges[self.curr_range][constr] = {"lower": 0, "upper": abs(value)}
        elif self.constraints[constr]["type"] == "L":
            self.ranges[self.curr_range][constr] = {"lower": -abs(value), "upper": 0}
        elif self.constraint[constr]["type"] == "E":
            if constr not in self.ranges[self.curr_range]:
                self.ranges[self.curr_range] = {"lower": 0, "upper": 0}
            if value < 0:
                self.ranges[self.curr_range][constr]["lower"] = value
            else:
                self.ranges[self.curr_range][constr]["upper"] = value
        else:
            raise ValueError('Range given for free row!')
    
    def _add_rhs_group(self, name):
        if name in self.rhs:
            raise ValueError('The RHS group ' + name + ' already exists!')
        self.rhs[name] = dict((constr,0) for constr in self.constraints)
        self.offsets[name] = dict((obj,0) for obj in self.objectives)
        if len(self.rhs) == 2:
            warnings.warn("There are multiple rhs groups for the linear program " + self.name + ". Switch rhs group using the MPS.attach_rhs function.", stacklevel=2)
    
    def attach_rhs(self, name):
        if name not in self.rhs:
            raise ValueError('The RHS group ' + name + ' does not exist!')
        self.curr_rhs = name
    
    def set_rhs(self, constr, value):
        if constr in self.constraints:
            self.rhs[self.curr_rhs][constr] = value
        elif constr in self.objectives:
            self.offsets[self.curr_rhs][constr] = value
        else:
            raise ValueError('The row ' + constr + ' does not exist!')
        
    def get_curr_rhs(self):
        return self.curr_rhs
    def get_curr_bnd(self):
        return self.curr_bnd
    def get_curr_range(self):
        return self.curr_range
    
    # used to produce dict
    def __iter__(self):
        yield ("name", self.name)
        yield ("objective_names", self.objective_names())
        yield ("bnd_names", self.bnd_names())
        yield ("rhs_names", self.rhs_names())
        yield ("range_names", self.range_names())
        yield ("constraint_names", self.constraint_names())
        yield ("variable_names", list(self._variables.keys()))
        #yield ("sos_names", self.sos_names())
        yield ("objectives", self.objectives)
        yield ("variables", self.variables)
        yield ("constraints", self.constraints)
        yield ("rhs", self.rhs)
        yield ("offsets", self.offsets)
        yield ("ranges", self.ranges)
        #yield ("sos", self.sos)


def read_mps(path, **kwargs):
    mode = -1
    
    default_bounds = {"c_lower": 0.0, "c_upper": math.inf, "i_lower": 0.0, "i_upper": math.inf}
    given_bounds = dict((k, kwargs[k]) for k in ['c_lower', 'c_upper', 'i_lower', 'i_upper'] if k in kwargs)
    default_bounds.update(given_bounds)
   
    mps = MPS(default_bounds)
    
    defaults = {"MI_upper": 0.0, "SC_lower": 1.0}
    given_defaults = dict((k, kwargs[k]) for k in ['MI_upper', 'SC_lower'] if k in kwargs)
    defaults.update(given_defaults)
    
    integral_marker = False
    
    with open(path, "r") as reader:
        
        for line in reader:
            if line.startswith("*"):
                continue

            line = re.split(" |\t", line)
            line = [x.strip() for x in line]
            line = list(filter(None, line))
            
            if len(line) == 0:
                continue
            if line[0] == "ENDATA":
                mps.finalize()
                break
            if line[0] == "*":
                continue
            if line[0] == "NAME":
                mps._set_name(line[1])
                #name = line[1]
                
            # ROW INSTRUCTION
            elif line[0] in [CORE_FILE_ROW_MODE, CORE_FILE_COL_MODE]:
                mode = MODE[line[0]]
                
            # RHS INSTRUCTION
            elif line[0] == CORE_FILE_RHS_MODE and len(line) <= 2:
                if len(line) > 1:
                    if line[1] not in mps.rhs:
                        mps._add_rhs_group(line[1])
                    mps.attach_rhs(line[1])
                    
                    mode = MODE[CORE_FILE_RHS_MODE_NAME_GIVEN]
                else:
                    mode = MODE[CORE_FILE_RHS_MODE_NO_NAME]
                    
            # BOUNDS INSTRUCTION  
            elif line[0] == CORE_FILE_BOUNDS_MODE and len(line) <= 2:
                if len(line) > 1:
                    if line[1] not in mps.variables:
                        mps._add_bnd_group(line[1])
                    mps.attach_bnd(line[1])
                    
                    mode = MODE[CORE_FILE_BOUNDS_MODE_NAME_GIVEN]
                else:
                    mode = MODE[CORE_FILE_BOUNDS_MODE_NO_NAME]
                    
            # RANGES INSTRUCTION
            elif line[0] == CORE_FILE_RANGES_MODE and len(line) <= 2:
                if len(line) > 1:
                    if line[1] not in mps.ranges:
                        mps._add_range_group(line[1])
                    mps.attach_range(line[1])
                    
                    mode = MODE[CORE_FILE_RANGES_MODE_NAME_GIVEN]
                else:
                    mode = MODE[CORE_FILE_RANGES_MODE_NO_NAME]
                    
            # SOS INSTRUCTION
            elif line[0] == CORE_FILE_SOS_MODE:
                mode = MODE[CORE_FILE_SOS_MODE_FIRST_LINE]
                
            
            elif line[0] == CORE_FILE_ARCS_MODE:
                raise ValueError("NETWORK is currently not implemented!")
                
                
            elif mode == MODE[CORE_FILE_ROW_MODE]:
                if line[0] == ROW_MODE_OBJ:
                    mps.add_objective(line[1])
                else:
                    mps._add_constraint(line[1], {"type": line[0], "coefficients": {}})
                    
            # COL MODE
            elif mode == MODE[CORE_FILE_COL_MODE]:
                if len(line) > 1 and line[1] == "'MARKER'":
                    if line[2] == "'INTORG'":
                        integral_marker = True
                    elif line[2] == "'INTEND'":
                        integral_marker = False
                    continue
                if line[0] not in mps._variables:
                    mps._add_variable(line[0], COL_TYPES[integral_marker])
                j = 1
                while j < len(line) - 1:
                    mps._set_coefficient(line[j], line[0], float(line[j+1]))
                    j = j + 2
                    
            # RHS MODE
            elif mode in [MODE[CORE_FILE_RHS_MODE_NAME_GIVEN], MODE[CORE_FILE_RHS_MODE_NO_NAME]]:
                if mode == MODE[CORE_FILE_RHS_MODE_NAME_GIVEN]:
                    if line[0] != mps.get_curr_rhs():
                        raise Exception("Other RHS name was given even though name was set after RHS tag.")
                elif line[0] not in mps.rhs:
                    mps._add_rhs_group(line[0])
                    mps.attach_rhs(line[0])
                for kk in range((len(line) - 1) // 2):
                    idx = kk * 2
                    mps.set_rhs(line[idx + 1], float(line[idx + 2]))
                    
            # SOS MODE
            elif mode == MODE[CORE_FILE_SOS_MODE_FIRST_LINE]:
                warnings.warn("SOS functionality is currently not supported. If your program relies on Special Ordered Set functionality the output might not be true to the file input!", stacklevel=2)
            #    degree = int(line[0][1:])
            #    mps.add_sos(line[2], degree, int(line[3]))
            #    mps.attach_sos(line[2])
                mode = MODE[CORE_FILE_SOS_MODE]
                
                
            elif mode == MODE[CORE_FILE_SOS_MODE]:
                pass
                #mps.add_variable_to_sos(line[1], float(line[2]))
                
            # BOUNDS MODE
            elif mode in [MODE[CORE_FILE_BOUNDS_MODE_NAME_GIVEN], MODE[CORE_FILE_BOUNDS_MODE_NO_NAME]]:
                if mode == MODE[CORE_FILE_BOUNDS_MODE_NAME_GIVEN]:
                    if line[1] != mps.get_curr_bnd():
                        raise Exception("Other BND name was given even though name was set after BND tag.")
                elif line[1] not in mps.variables:
                    mps._add_bnd_group(line[1])
                    mps.attach_bnd(line[1])
                
                if line[0] == "UP":
                    mps._update_variable(line[2], {"upper": float(line[3])})
                elif line[0] == "UI":
                    mps._update_variable(line[2], {"type": "Integer", "upper": float(line[3])})
                elif line[0] == "LO":
                    mps._update_variable(line[2], {"lower": float(line[3])})
                elif line[0] == "LI":
                    mps._update_variable(line[2], {"type": "Integer", "lower": float(line[3])})
                elif line[0] == "FX":
                    mps._update_variable(line[2], {"lower": float(line[3]), "upper": float(line[3])})
                elif line[0] == "FR":
                    mps._update_variable(line[2], {"lower": -math.inf, "upper": math.inf})
                elif line[0] == "MI":
                    mps._update_variable(line[2], {"lower": -math.inf, "upper": defaults["MI_upper"]})
                elif line[0] == "PL":
                    mps._update_variable(line[2], {"lower": 0.0, "upper": math.inf})
                elif line[0] == "BV":
                    mps._update_variable(line[2], {"type": "Integer", "lower": 0.0, "upper": 1.0})
                elif line[0] == "SC":
                    val = defaults["SC_lower"]
                    if len(line) > 3:
                        val = float(line[3])
                    mps._update_variable(line[2], {"type": "Semi-Continuous", "lower": val})
                
            # RANGES MODE
            elif mode in [MODE[CORE_FILE_RANGES_MODE_NAME_GIVEN], MODE[CORE_FILE_RANGES_MODE_NO_NAME]]:
                if mode == MODE[CORE_FILE_RANGES_MODE_NAME_GIVEN]:
                    if line[0] != mps.get_curr_range():
                        raise Exception("Other RANGES group tag was given even though group was set after RANGES tag.")
                elif line[0] not in mps.ranges:
                    mps._add_range_group(line[0])
                    mps.attach_range(line[0])
                
                mps.add_range(line[1], float(line[2]))
    
    
    return mps


# Copyright (c) 2024, InfinityQ Technology, Inc.

from typing import Union
from titanq._model.variable import Expression, VariableVector, VectorExpression


def fastSum(exprVector: Union[VectorExpression, VariableVector]) -> Expression:
    """
    ℹ️ **This feature is experimental and may change.**

    Computes the sum of all elements in the provided `VectorExpression` or `VariableVector`.
    This function is an faster alternative to the traditional `sum()` operation for vector-based expressions.

    Parameters
    ----------
    exprVector
        The `VectorExpression` or `VariableVector` whose elements are to be summed.

    Returns
    -------
    Expression
        An `Expression` representing the sum of the elements in `exprVector`.

    Raises
    ------
    ValueError
        If the provided input is not of type `VectorExpression` or `VariableVector`.

    Examples
    --------
    >>> from titanq import Model, Vtype
    >>> x = model.add_variable_vector('x', 1000, Vtype.BINARY)
    >>> y = model.add_variable_vector('y', 1000, Vtype.BINARY))
    >>> exprA = fastSum((np.array([3, 4]) * x + (x * y) - 5 * y))
    >>> exprB = fastSum(x)
    """
    if not isinstance(exprVector, (VectorExpression, VariableVector)):
        raise ValueError(f"invalid input of type {type(expression).__name__}")
    
    return exprVector._fastSum()



# Copyright (c) 2024, InfinityQ Technology, Inc.

from pathlib import Path
import numpy as np

from titanq.tools._mps_loader import MPS, read_mps

from .._model.errors import MPSFileError
from .._model.model import Model
from .._model.variable import Vtype


def configure_model_from_mps_file(model: Model, file_path: Path) -> None:
    """
    Configure a model with an MPS file. Set the variable vector, the
    objective matrices and the constraints.

    Parameters
    ----------
    model
        The instance of the model to configure.
    file_path
        The path to the MPS file.
    """
    try:
        mps: MPS = read_mps(file_path)

        variables = mps.get_variables()
        var_list = list(variables.keys())
        nbr_of_variables = len(variables)
        coefficients = np.array([mps.get_objectives()[mps.objective_names()[0]]['coefficients'].get(var, 0) for var in var_list])

        for var_name, var_info in variables.items():
            var_type = var_info['type']
            var_lower = var_info['lower']
            var_upper = var_info['upper']
            if var_type == 'Integer' and var_lower == 0 and var_upper == 1:
                model.add_variable_vector(var_name, 1, Vtype.BINARY)
            elif var_type == 'Integer':
                model.add_variable_vector(var_name, 1, Vtype.INTEGER, [(var_lower, var_upper)])
            else:
                model.add_variable_vector(var_name, 1, Vtype.CONTINUOUS, [(var_lower, var_upper)])

        model.set_objective_matrices(np.zeros((nbr_of_variables, nbr_of_variables), dtype=np.float32), coefficients.astype(np.float32))

        for constr, constr_value in mps.get_constraints().items():
            constraint_mask = np.array([constr_value['coefficients'].get(var, 0) for var in var_list])
            rhs = mps.get_rhs()[constr]
            if constr_value['type'] == 'G':
                constraint_bounds = np.array([rhs, np.nan])
                if mps.get_ranges().get(constr) is not None:
                    constraint_bounds[1] = rhs+mps.get_ranges().get(constr)['upper']
                model.add_inequality_constraint(constraint_mask.astype(np.float32), constraint_bounds.astype(np.float32))

            elif constr_value['type'] == 'L':
                constraint_bounds = np.array([np.nan, rhs])
                if mps.get_ranges().get(constr) is not None:
                    constraint_bounds[0] = rhs+mps.get_ranges().get(constr)['lower']
                model.add_inequality_constraint(constraint_mask.astype(np.float32), constraint_bounds.astype(np.float32))

            elif constr_value['type'] == 'E':
                model.add_equality_constraint(constraint_mask.astype(np.float32), mps.get_rhs().get(constr, 0))

    except Exception as e:
        raise MPSFileError(str(e))


# Copyright (c) 2024, InfinityQ Technology, Inc.
from typing import Dict, List, Set

from .variable import VariableVector, Vtype
from .errors import VariableAlreadyExist

import numpy as np
from numpy._typing import NDArray

class VariableVectorList:
    def __init__(self) -> None:
        self._variables: List[VariableVector] = []
        self._variables_name: Set[str] = set()


    def add(self, v: VariableVector):
        """
        Add a variable vector to this list

        :param v: the new variable vector to add

        :raise VariableAlreadyExist: If a the variable to be add has the same name as an existing variable in this list
        """
        name = v.name()
        if name in self._variables_name:
            raise VariableAlreadyExist(f"variable with name: {name} has already been defined")

        self._variables_name.add(name)
        self._variables.append(v)


    def total_variable_size(self) -> int:
        """ :return: total size of all variables vector contained in the list """
        return sum(len(v) for v in self._variables)


    def n_variables(self) -> int:
        """ :return: The number of variable vector in the list """
        return len(self._variables)


    def get_variable_types_str(self) -> str:
        """
        Generate the variable types string for all the variable of this list.

        :return: The generated variable type string for all the variables
        """

        return "".join(v.variable_types_as_list() for v in self._variables)


    def has_variable_of_type(self, vtype: Vtype) -> bool:
        """:return: If any of the variable in this list has the given type"""
        return any(v.vtype() is vtype for v in self._variables)


    def variable_bounds(self) -> NDArray:
        """return the variable bounds of all variable in this list"""
        return np.concatenate([v.variable_bounds() for v in self._variables])


    def index_variables(self) -> Dict[VariableVector, int]:
        """return a dictionnary of all variable vectors with their start index in the list"""
        variables_index = {}
        current_index = 0
        for var in self._variables:
            variables_index[var.name()] = current_index
            current_index += len(var)
        return variables_index


    def __iter__(self):
        return iter(self._variables)



# Copyright (c) 2024, InfinityQ Technology, Inc.

class Manifest:
    """
    This class tracks gives us more insignt on the different numpy data used.
    """
    def __init__(self) -> None:
        self._has_cardinality_constraint: bool = False
        self._has_set_partitioning_constraint: bool = False
        self._has_equality_constraint: bool = False
        self._has_inequality_constraint: bool = False

    def activate_set_partitioning_constraint(self) -> None:
        """
        Declares set partitioning constraint is being used.
        """
        self._has_set_partitioning_constraint = True

    def activate_cardinality_constraint(self) -> None:
        """
        Declares cardinality constraint is being used.
        """
        self._has_cardinality_constraint = True

    def activate_equality_constraint(self) -> None:
        """
        Declares equality constraint is being used.
        """
        self._has_equality_constraint = True

    def activate_inequality_constraint(self) -> None:
        """
        Declares inequality constraint is being used.
        """
        self._has_inequality_constraint = True

    def has_set_partitioning_constraint(self) -> bool:
        """
        :return: if the manifest has set partitioning constraint.
        """
        return self._has_set_partitioning_constraint

    def has_cardinality_constraint(self) -> bool:
        """
        :return: if the manifest has cardinality constraint.
        """
        return self._has_cardinality_constraint

    def has_equality_constraint(self) -> bool:
        """
        :return: if the manifest has equality constraint.
        """
        return self._has_equality_constraint

    def has_inequality_constraint(self) -> bool:
        """
        :return: if the manifest has inequality constraint.
        """
        return self._has_inequality_constraint



# Copyright (c) 2024, InfinityQ Technology, Inc.
import logging
from typing import Optional, Type
import numpy as np
import requests
import time
import urllib3

from .._client.model import UrlInput, UrlOutput
from .storage_client import StorageClient, TempFileManager
from .._client.client import Client
from .._model.errors import ConnectionError

log = logging.getLogger("TitanQ")


class ManagedStorage(StorageClient):
    def __init__(self, titanq_client: Client):
        """
        Initiate the managed storage client for handling the temporary files.

        :titanq_client: titanq_client to be used to fetch temporary URL's
        """
        self._titanq_client = titanq_client
        self._urls = None
        self._weights_uploaded = False
        self._constraint_bounds_uploaded = False
        self._constraint_weights_uploaded = False
        self._variable_bounds_uploaded = False

    def temp_files_manager(self,
        bias: np.ndarray,
        weights: np.ndarray,
        constraint_bounds: Optional[np.ndarray],
        constraint_weights: Optional[np.ndarray],
        variable_bounds: Optional[np.ndarray]
    ) -> TempFileManager:
        """
        We override temp_files_manager to fetch new urls every time we solve.(resetting temp storage timeout)
        """
        self._urls = self._titanq_client.temp_storage()
        return super().temp_files_manager(bias, weights, constraint_bounds, constraint_weights, variable_bounds)

    def _upload_arrays(
        self,
        bias: bytes,
        weights: Optional[bytes],
        constraint_bounds: Optional[bytes],
        constraint_weights: Optional[bytes],
        variable_bounds: Optional[bytes]
        ):
        upload_tuple = [(self._urls.input.bias_file.upload, bias)]

        if weights:
            self._weights_uploaded = True
            upload_tuple.append((self._urls.input.weights_file.upload, weights))

        if constraint_bounds:
            self._constraint_bounds_uploaded = True
            upload_tuple.append((self._urls.input.constraint_bounds_file.upload, constraint_bounds))

        if constraint_weights:
            self._constraint_weights_uploaded = True
            upload_tuple.append((self._urls.input.constraint_weights_file.upload, constraint_weights))

        if variable_bounds:
            self._variable_bounds_uploaded = True
            upload_tuple.append((self._urls.input.variable_bounds_file.upload, variable_bounds))

        log.debug(f"Uploading files on our temporary storage")
        for url, data in upload_tuple:
            requests.put(url, data=data)


    def _input(self) -> UrlInput:
        return UrlInput(
            bias_file_name=self._urls.input.bias_file.download,
            weights_file_name=self._urls.input.weights_file.download if self._weights_uploaded else None,
            constraint_bounds_file_name=self._urls.input.constraint_bounds_file.download if self._constraint_bounds_uploaded else None,
            constraint_weights_file_name=self._urls.input.constraint_weights_file.download if self._constraint_weights_uploaded else None,
            variable_bounds_file_name=self._urls.input.variable_bounds_file.download if self._variable_bounds_uploaded else None,
            manifest=None
        )

    def _output(self) -> UrlOutput:
        return UrlOutput(result_archive_file_name=self._urls.output.result_archive_file.upload)

    def _wait_for_result_to_be_uploaded_and_download(self) -> bytes:
        """
        Wait until the content of the file in the temporary storage is bigger
        than 0 bytes. Meaning it will wait until the file is uploaded

        :param url: Url to download the file.
        """
        retries = 0

        while retries < 5:
            try:
                response = requests.get(self._urls.output.result_archive_file.download)
                response.raise_for_status()
                if len(response.content) > 0:
                    return response.content

            except (urllib3.exceptions.ProtocolError, requests.exceptions.ConnectionError, ConnectionResetError) as e:
                retries = retries + 1
                log.warning(f"Caught error {e} [retries: {retries}]")
                time.sleep(1.0)

            time.sleep(0.25)

        raise ConnectionError("Unexpected error with InfinityQ internal storage, please contact InfinityQ support for more information")

    def _delete_remote_object(self) -> None:
        log.debug("Temporary storage option does not delete any file at the moment")



# Copyright (c) 2024, InfinityQ Technology, Inc.
from . import errors
from .model import Model
from .objective import Target
from .optimize_response import OptimizeResponse
from .variable import Vtype
from .precision import Precision
from .fastSum import fastSum



# Copyright (c) 2024, InfinityQ Technology, Inc.

import io
import json
import logging
import numpy as np
import os
import zipfile
from typing import Any, Dict, List, Optional, Tuple, Union
from warnings import warn

from .constraints import Constraints
from .errors import (
    ConstraintAlreadySetError,
    MissingObjectiveError,
    MissingVariableError,
    MissingTitanqApiKey,
    ObjectiveAlreadySetError,
    OptimizeError,
)
from .expression_parser import process_expression
from .manifest import Manifest
from .objective import Objective, Target
from .optimize_response import OptimizeResponse
from .precision import Precision
from .variable import (
    BinaryVariableVector,
    ConstraintExpression,
    ConstraintType,
    ContinuousVariableVector,
    Expression,
    IntegerVariableVector,
    VariableVector,
    Vtype,
)
from .._client import api_model, Client
from .._storage import ManagedStorage, StorageClient
from .numpy_util import (
    is_ndarray_binary,
    is_upperbound_bigger_equal_lowerbound,
    ndarray_contains_nan_inf,
    reshape_to_2d,
    validate_cardinalities
)
from .variable_list import VariableVectorList

log = logging.getLogger("TitanQ")


class Model:
    """
    Root object to define a problem to be optimized
    """

    def __init__(
        self,
        *,
        api_key: str = None,
        storage_client: StorageClient = None,
        base_server_url: str = "https://titanq.infinityq.io"
        ) -> None:
        """
        Initiate the model with a storage client. If the storage_client is missing, the storage will be managed by TitanQ.

        Notes
        -----
        The storage managed by TitanQ supports weight matrices with a size up to 10k only.

        Parameters
        ----------
        api_key
            TitanQ API key to access the service.
            If not set, it will use the environment variable ``TITANQ_API_KEY``
        storage_client
            Storage to choose in order to store some items.
        base_server_url
            TitanQ API server url, default set to ``https://titanq.infinityq.io``.

        Raises
        ------
        MissingTitanqApiKey
            If no API key is set and is also not set as an environment variable

        Examples
        --------
        With an S3 storage client
            >>> from titanq import Model, S3Storage
            >>> storage_client = S3Storage(
                access_key="{insert aws bucket access key here}",
                secret_key="{insert aws bucket secret key here}",
                bucket_name="{insert bucket name here}"
            )
            >>> model = Model(storage_client)

        Managed storage client
            >>> from titanq import Model, S3Storage
            >>> model = Model()
        """
        self._variables = VariableVectorList()
        self._objective: Objective = None
        self._constraints = Constraints()

        api_key = api_key or os.getenv("TITANQ_API_KEY")
        if api_key is None:
            raise MissingTitanqApiKey(
                "No API key is provided. You can set your API key in the Model, "
                + "or you can set the environment variable TITANQ_API_KEY")

        self._titanq_client = Client(api_key, base_server_url)
        self._manifest: Manifest = Manifest()

        # the user chose a managed storage or left it as default
        if storage_client is None:
            storage_client = ManagedStorage(self._titanq_client)

        self._storage_client = storage_client


    def get_objective_matrices(self) -> Tuple[Optional[np.ndarray], Optional[np.ndarray]]:
        """
        Retrieve the weights and bias vector from the model's objective. Both will be None
        if not set.

        Return
        ------
        Weights matrix if not None and bias vector if not None
        """
        weights = self._objective.weights() if self._objective else None
        bias = self._objective.bias() if self._objective else None
        return (weights, bias)

    def get_constraints_weights_and_bounds(self) -> Tuple[Optional[np.ndarray], Optional[np.ndarray]]:
        """
        Retrieve the weights and bounds of all constraints from the model.

        Return
        ------
        Constraints weights if not None and bounds if not None
        """
        return self._constraints.weights(), self._constraints.bounds()


    def add_variable_vector(
        self,
        name: str = '',
        size: int = 1,
        vtype: Vtype = Vtype.BINARY,
        variable_bounds:
            Optional[
                Union[
                    List[Tuple[int, int]],
                    List[Tuple[float, float]],
                ]
            ]=None,
    ) -> VariableVector:
        """
        Add a vector of variable to the model. Multiple variables vector can be added but with different names.

        Notes
        -----
        If Vtype is set to ``Vtype.INTEGER`` or ``Vtype.CONTINUOUS``, variable_bounds need to be set.

        Parameters
        ----------
        name
            The name given to this variable vector.
        size
            The size of the vector.
        vtype
            Type of the variables inside the vector.
        variable_bounds
            Lower and upper bounds for the variable vector. A list of tuples (can be either integers or continuous)

        Return
        ------
        variable
            The variable vector created.

        Raises
        ------
        MaximumVariableLimitError
            If the total size of variables exceed the limit.
        ValueError
            If the size of the vector is < 1

        Examples
        --------
        >>> from titanq import Model, Vtype
        >>> model.add_variable_vector('x', 3, Vtype.BINARY)
        >>> model.add_variable_vector('y', 2, Vtype.INTEGER, [[0, 5], [1, 6]])
        >>> model.add_variable_vector('z', 3, Vtype.CONTINUOUS, [[2.3, 4.6], [3.1, 5.3], [1.1, 4]])
        """

        if not variable_bounds:
            variable_bounds = []

        # validation
        if not self._constraints.is_empty():
            raise ConstraintAlreadySetError("Cannot add additional variable once constraints have been defined")

        if self._objective is not None:
            raise ObjectiveAlreadySetError("Cannot add additional variable once objective have been defined")

        # create the Variable vector
        if vtype is Vtype.BINARY:
            if variable_bounds:
                raise ValueError("variable_bounds is not supported with Vtype.BINARY")
            variables = BinaryVariableVector(name, size)

        elif vtype is Vtype.INTEGER:
            variables = IntegerVariableVector(name, size, variable_bounds)

        elif vtype is Vtype.CONTINUOUS:
            variables = ContinuousVariableVector(name, size, variable_bounds)

        else:
            raise NotImplementedError(f"Unsupported variable type: {vtype}")

        self._variables.add(variables)
        self._constraints.augment_size(size)

        log.debug(f"add variable name='{name}', type={str(vtype)}, size={size}.")

        return variables

    def set_objective_expression(self, expr: Expression, target=Target.MINIMIZE):
        """
        ℹ️ **This feature is experimental and may change.**

        Sets the objective function for the optimization problem using the given expression.

        This method processes the provided expression to extract the bias vector and weight matrix,
        and then sets these as the objective matrices for the optimization problem.

        Parameters
        ----------
        expr
            The expression defining the objective function. This should be an instance of `Expression`.
        target
            The target of this objective matrix.

        Raises
        ------
        TypeError
            if the provided expression contains any invalid/unsupported input

        Examples
        --------
        >>> from titanq import Model, Vtype
        >>> x = model.add_variable_vector('x', 2, Vtype.BINARY)
        >>> y = model.add_variable_vector('y', 2, Vtype.BINARY)
        >>> expr = (np.array([3, 4]) * x + (x * y) - 5 * y)[0]
        >>> model.set_objective_expression(expr)
        """
        weights, bias, _ = process_expression(expr, self._variables)
        self.set_objective_matrices(weights, bias, target)


    def add_constraint_from_expression(self, expression: ConstraintExpression, inequalityTolerance: float=0):
        """
        ℹ️ **This feature is experimental and may change.**

        Adds a constraint to the model using the given expression.

        This method processes the provided constraint expression to add it as a constraint to the 
        optimization problem. Only linear constraints of the following types are supported:

        - `A == B`

        - `A < B`
        
        - `A <= B`
        
        - `A > B`
        
        - `A >= B`
        
        Constraints involving quadratic terms are not supported and will raise an error.

        Parameters
        ----------
        expression
            The constraint expression. This should be an instance of `ConstraintExpression`.
        inequalityTolerance
            A unsigned float specifying the tolerance level for the inequality constraint.
            A strict inequality such as `x < 3` will be interpreted as `x <= 3 - inequalityTolerance`

        Raises
        ------
        ValueError
            If the provided expression contains quadratic terms or if `inequalityTolerance` is set to a negative value.
        TypeError
            If the provided expression is of an invalid or unsupported type.

        Examples
        --------
        >>> from titanq import Model, Vtype
        >>> x = model.add_variable_vector('x', 2, Vtype.BINARY)
        >>> y = model.add_variable_vector('y', 2, Vtype.BINARY)
        >>> expr = fastSum(x+y) == 1
        >>> model.add_constraint_from_expression(expr)
        """
        if not inequalityTolerance >= 0:
            raise ValueError("InequalityTolerance is not an unsigned float") 
        
        if not isinstance(expression, ConstraintExpression):
            raise TypeError(f"invalid input of type {type(expression).__name__}")
        
        weights, mask, limit = process_expression(expression.expression(), self._variables)
        if weights is not None:
            raise ValueError(
                "Quadratic terms are not supported in constraints. "
                "Please ensure that the expression contains only linear terms when using it in a constraint."
            )
           
        constraintType =expression.constraintType()
        if constraintType == ConstraintType.EQUAL:
            self.add_equality_constraint(mask, limit)
        elif constraintType == ConstraintType.GREATER:
            self.add_inequality_constraint(mask, np.array([limit+inequalityTolerance, np.nan], dtype=np.float32))
        elif constraintType == ConstraintType.GREATER_EQUAL:
            self.add_inequality_constraint(mask, np.array([limit, np.nan], dtype=np.float32))
        elif constraintType == ConstraintType.LESSER:
            self.add_inequality_constraint(mask, np.array([np.nan, limit-inequalityTolerance], dtype=np.float32))
        elif constraintType == ConstraintType.LESSER_EQUAL:
            self.add_inequality_constraint(mask, np.array([np.nan, limit], dtype=np.float32))
        

    def set_objective_matrices(self, weights: Optional[np.ndarray], bias: np.ndarray, target=Target.MINIMIZE):
        """
        Set the objective matrices for the model.

        Parameters
        ----------
        weights
            The quadratic objective matrix, **this matrix needs to be symmetrical**.
            A NumPy 2-D dense ndarray (must be float32).
            Weights matrix can be set to **None** if it is a linear problem with no quadratic elements.
        bias
            The linear constraint vector. A NumPy 1-D ndarray.
        target
            The target of this objective matrix.

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        ObjectiveAlreadySetError
            If an objective has already been set in this model.
        ValueError
            If the weights shape or the bias shape does not fit the variables in the model.
            If the weights or bias data type is not float32.

        Examples
        --------
        >>> from titanq import Model, Target
        >>> edges = {0:[4,5,6,7], 1:[4,5,6,7], 2:[4,5,6,7], 3:[4,5,6,7], 4:[0,1,2,3], 5:[0,1,2,3], 6:[0,1,2,3], 7:[0,1,2,3]}
        >>> size = len(edges)
        >>> weights = np.zeros((size, size), dtype=np.float32)
        >>> for root, connections in edges.items():
        >>>     for c in connections:
        >>>         weights[root][c] = 1
        >>> # construct the bias vector (Uniform weighting across all nodes)
        >>> bias = np.asarray([0]*size, dtype=np.float32)
        >>> model.set_objective_matrices(weights, bias, Target.MINIMIZE)
        """
        if self._variables.n_variables() == 0:
            raise MissingVariableError("Cannot set objective before adding a variable to the model.")

        if self._objective is not None:
            raise ObjectiveAlreadySetError("An objective has already have been set for this model.")

        log.debug(f"set objective matrix and bias vector.")

        self._objective = Objective(self._variables.total_variable_size(), weights, bias, target)


    def add_set_partitioning_constraints_matrix(self, constraint_mask: np.ndarray):
        """
        Adds set partitioning constraints in matrix format to the model.

        Parameters
        ----------
        constraint_mask
            A NumPy 2-D dense ndarray (must be binary).
            The constraint_mask matrix of shape (M, N) where M the number of constraints and N is the number of variables.

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        MaximumConstraintLimitError
            The number of constraints exceed the limit.
        ConstraintSizeError
            If the constraint_mask shape does not fit the expected shape of this model.
        ValueError
            If the constraint_mask data type is not binary.

        Examples
        --------
        >>> constraint_mask = np.array([[1, 1, 1, 0, 1], [1, 1, 1, 1, 0]])
        >>> model.add_set_partitioning_constraints_matrix(constraint_mask)
        """
        if self._variables.n_variables() == 0:
            raise MissingVariableError("Cannot set constraints before adding a variable to the model.")

        if constraint_mask.ndim == 1:
            raise ValueError(
                "Cannot use add_set_partitioning_constraints_matrix() function with a vector, " \
                "please use add_set_partitioning_constraint() instead")

        if not is_ndarray_binary(constraint_mask):
            raise ValueError(f"Cannot add a constraint if the values are not in binary.")

        self._constraints.add_constraint(
            constraint_weights=constraint_mask,
            constraint_bounds=np.ones((constraint_mask.shape[0],2))
        )
        self._manifest.activate_set_partitioning_constraint()

    def add_set_partitioning_constraint(self, constraint_mask: np.ndarray):
        """
        Adds set partitioning constraint vector to the model.

        Parameters
        ----------
        constraint_mask
            A NumPy 1-D dense ndarray (must be binary).
            The constraint_mask vector of shape (N,) where N is the number of variables.

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        MaximumConstraintLimitError
            The number of constraint exceed the limit.
        ConstraintSizeError
            If the constraint_mask shape does not fit the expected shape of this model.
        ValueError
            If the constraint_mask data type is not binary.

        Examples
        --------
        >>> constraint_mask = np.array([1, 1, 1, 0, 1])
        >>> model.add_set_partitioning_constraint(constraint_mask)
        """
        if constraint_mask.ndim > 1:
            raise ValueError(
                "Cannot use this add_set_partitioning_constraint() function with a matrix, " \
                "please use add_set_partitioning_constraints_matrix() instead")

        self.add_set_partitioning_constraints_matrix(reshape_to_2d(constraint_mask))


    def add_cardinality_constraints_matrix(self, constraint_mask: np.ndarray, cardinalities: np.ndarray):
        """
        Adds cardinality constraints in matrix format to the model.

        Parameters
        ----------
        constraint_mask
            A NumPy 2-D dense ndarray (must be binary).
            The constraint_mask matrix of shape (M, N) where M the number of constraints and N is the number of variables.
        cardinalities
            A NumPy 1-D ndarray (must be non-zero unsigned integer).
            The constraint_rhs vector of shape (M,) where M is the number of constraints.

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        MaximumConstraintLimitError
            The number of constraint exceed the limit.
        ConstraintSizeError
            If the constraint_mask shape or the constraint_rhs shape does not fit the expected shape of this model.
        ValueError
            If the constraint_mask is not binary or cardinalities data type are not unsigned integers.

        Examples
        --------
        >>> constraint_mask = np.array([[1, 1, 1, 0, 1], [1, 1, 1, 1, 0]])
        >>> cardinalities = np.array([3, 2])
        >>> model.add_cardinality_constraints_matrix(constraint_mask, cardinalities)
        """
        if self._variables.n_variables() == 0:
            raise MissingVariableError("Cannot set constraints before adding a variable to the model.")

        if constraint_mask.ndim == 1:
            raise ValueError(
                "Cannot use add_cardinality_constraints_matrix() function with a vector, " \
                "please use add_cardinality_constraint() instead")

        if cardinalities.ndim != 1:
            raise ValueError(f"Cannot set constraints if cardinalities is not a NumPy 1-D dense ndarray")

        if not np.issubdtype(cardinalities.dtype, np.integer):
            raise ValueError("Found cardinalities data types not integer")

        if not np.all(cardinalities > 0):
            raise ValueError("Found cardinalities data types not unsigned integer")

        if not is_ndarray_binary(constraint_mask):
            raise ValueError(f"Cannot add a constraint if the values are not in binary.")

        if cardinalities.shape[0] != constraint_mask.shape[0]:
            raise ValueError(
                f"Cannot set constraints if cardinalities shape is not the same as the expected shape of this model." \
                f" Got cardinalities shape: {cardinalities.shape[1]}, constraint mask shape: {constraint_mask.size()}.")

        validate_cardinalities(constraint_mask, cardinalities)

        self._constraints.add_constraint(
            constraint_weights=constraint_mask,
            constraint_bounds=np.repeat(cardinalities, 2).reshape(-1, 2)
        )
        self._manifest.activate_cardinality_constraint()


    def add_cardinality_constraint(self, constraint_mask: np.ndarray, cardinality: int):
        """
        Adds cardinality constraint vector to the model.

        Parameters
        ----------
        constraint_mask
            A NumPy 1-D dense ndarray (must be binary).
            The constraint_mask vector of shape (N,) where N is the number of variables.
        cardinality
            The constraint_rhs cardinality.
            This value has to be a non-zero unsigned integer.

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        MaximumConstraintLimitError
            If the number of constraints exceed the limit.
        ConstraintSizeError
            If the constraint_mask shape or the constraint_rhs shape does not fit
            the expected shape of this model.
        ValueError
            If the constraint_mask is not in binary or the cardinality is not an unsigned integer.

        Examples
        --------
        >>> constraint_mask = np.array([1, 1, 1, 0, 1])
        >>> cardinality = 3
        >>> model.add_cardinality_constraint(constraint_mask, cardinality)
        """
        if constraint_mask.ndim > 1:
            raise ValueError(
                "Cannot use add_cardinality_constraint() function with a matrix, " \
                "please use add_cardinality_constraints_matrix() instead")

        self.add_cardinality_constraints_matrix(reshape_to_2d(constraint_mask), np.full((1,), cardinality))


    def add_equality_constraints_matrix(self, constraint_mask: np.ndarray, limit: np.ndarray) -> None:
        """
        Adds an equality constraint matrix to the model.

        Parameters
        ----------
        constraint_mask
            A NumPy 2-D dense ndarray (float32).
            The constraint_mask vector of shape (M, N) where M the number of constraints and N is the number of variables.
        limit
            A NumPy 1-D array (float32).
            The limit vector of shape (M,) where M is the number of constraints.

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        MaximumConstraintLimitError
            The number of constraint exceed the limit.
        ValueError
            If the constraint_mask shape does not fit the expected shape of this model.
            If the constraint_mask or limit contains irregular format ('NaN' or 'inf').

        Examples
        --------
        >>> constraint_mask = np.array([[-3.51, 0, 0, 0], [10, 0, 0, 0]], dtype=np.float32)
        >>> limit = np.array([2, 10], dtype=np.float32)
        >>> model.add_equality_constraints_matrix(constraint_mask, limit)
        """
        if self._variables.n_variables() == 0:
            raise MissingVariableError("Cannot set constraints before adding a variable to the model.")

        if constraint_mask.dtype != np.float32 or limit.dtype != np.float32:
            raise ValueError(f"Input parameters must be float32, got Constraint mask: {constraint_mask.dtype}, Limit: {limit.dtype}")

        if constraint_mask.ndim == 1:
            raise ValueError(
                "Cannot use add_equality_constraint_matrix() function with a vector, " \
                "please use add_equality_constraint() instead")

        if ndarray_contains_nan_inf(constraint_mask):
            raise ValueError("Constraint mask contains NaN  or inf values")

        if limit.ndim != 1:
            raise ValueError("Limit must be a NumPy 1-D array")

        if ndarray_contains_nan_inf(limit):
            raise ValueError("Limit contains NaN or inf values")

        self._constraints.add_constraint(
            constraint_weights=constraint_mask,
            constraint_bounds=np.repeat(limit, 2).reshape(-1, 2)
        )
        self._manifest.activate_equality_constraint()


    def add_equality_constraint(self, constraint_mask: np.ndarray, limit: np.float32) -> None:
        """
        Adds an equality constraint vector to the model.

        Parameters
        ----------
        constraint_mask
            A NumPy 1-D dense ndarray (float32).
            The constraint_mask vector of shape (N,) where N is the number of variables.
        limit
            Limit value to the constraint mask.

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        MaximumConstraintLimitError
            The number of constraint exceed the limit.
        ValueError
            If the constraint_mask shape does not fit the expected shape of this model.
            If the constraint_mask or limit contains irregular format ('NaN' or 'inf').

        Examples
        --------
        >>> constraint_mask = np.array([1.05, -1.1], dtype=np.float32)
        >>> limit = -3.45
        >>> model.add_equality_constraint(constraint_mask, limit)
        """
        if constraint_mask.ndim > 1:
            raise ValueError(
                "Cannot use add_equality_constraint() function with a matrix, " \
                "please use add_equality_constraint_matrix() instead")

        self.add_equality_constraints_matrix(reshape_to_2d(constraint_mask), np.full((1,), limit, dtype=np.float32))


    def add_inequality_constraints_matrix(self, constraint_mask: np.ndarray, constraint_bounds: np.ndarray):
        """
        Adds inequality constraint matrix to the model.

        Parameters
        ----------
        constraint_mask
            A NumPy 2-D dense ndarray (float32).
            The constraint_mask vector of shape (M, N) where N is the number of variables.
        constraint_bounds
            A NumPy 2-D ndarray (float32).
            Vector of shape (M, 2) where M is the number of constraints.

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        MaximumConstraintLimitError
            The number of constraint exceed the limit.
        ValueError
            If the constraint_mask shape does not fit the expected shape of this model.
            If the constraint_mask contains irregular format ('NaN' or 'inf').
            If the lowerbound is equal or higher than its given upperbound.

        Examples
        --------
        >>> constraint_mask = np.array([[-3.51, 0], [10, 0]], dtype=np.float32)
        >>> constraint_bounds = np.array([[8, 9], [np.nan, 100_000]], dtype=np.float32)
        >>> model.add_inequality_constraints_matrix(constraint_mask, constraint_bounds)
        """
        if self._variables.n_variables() == 0:
            raise MissingVariableError("Cannot set constraints before adding a variable to the model.")

        if constraint_mask.dtype != np.float32 or constraint_bounds.dtype != np.float32:
            raise ValueError(f"Input parameters must be float32, got Constraint mask: {constraint_mask.dtype}, Limit: {constraint_bounds.dtype}")

        if constraint_mask.ndim == 1:
            raise ValueError(
                "Cannot use add_inequality_constraint_matrix() function with a vector, " \
                "please use add_inequality_constraint() instead")

        if ndarray_contains_nan_inf(constraint_mask):
            raise ValueError("Constraint mask contains NaN  or inf values.")

        if is_upperbound_bigger_equal_lowerbound(constraint_bounds):
            raise ValueError("Constraint bounds contains lowerbounds equal or higher than their upperbound.")

        self._constraints.add_constraint(
            constraint_weights=constraint_mask,
            constraint_bounds=constraint_bounds
        )
        self._manifest.activate_inequality_constraint()


    def add_inequality_constraint(self, constraint_mask: np.ndarray, constraint_bounds: np.ndarray):
        """
        Adds inequality constraint vector to the model. At least one bound must be set.

        Parameters
        ----------
        constraint_mask
            A NumPy 1-D dense ndarray (float32).
            The constraint_mask vector of shape (N,) where N is the number of variables.
        constraint_bounds
            A NumPy 1-D ndarray (float32).
            Vector of shape (2,)

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        MaximumConstraintLimitError
            The number of constraint exceed the limit.
        ValueError
            If the constraint_mask shape does not fit the expected shape of this model.
            If the constraint_mask contains irregular format ('NaN' or 'inf').
            If the lowerbound is equal or higher than the upperbound.

        Examples
        --------
        >>> constraint_mask = np.array([1.05, -1.1], dtype=np.float32)
        >>> constraint_bounds = np.array([1.0, np.nan], dtype=np.float32)
        >>> model.add_inequality_constraint(constraint_mask, constraint_bounds)
        """
        if constraint_mask.ndim > 1:
            raise ValueError(
                "Cannot use add_inequality_constraint() function with a matrix, " \
                "please use add_inequality_constraint_matrix() instead")

        self.add_inequality_constraints_matrix(reshape_to_2d(constraint_mask), reshape_to_2d(constraint_bounds))

    def optimize(
        self,
        *,
        beta: List[float] = [1, 0.5, 0.33, 0.25, 0.2, 0.16, 0.14, 0.125],
        coupling_mult: float = 0.5,
        timeout_in_secs: float = 10.0,
        num_chains: int = 8,
        num_engines: int = 1,
        penalty_scaling: float = None,
        precision: Precision = Precision.AUTO,
    ) -> OptimizeResponse:
        """
        Optimize this model.

        Notes
        -----
        All of the files used during this computation will be cleaned at the end.
        For more information on how to tunes those parameters, visit

        `The tuning guide <https://docs.titanq.infinityq.io/user-guide/parameter_tuning_guide>`_

        `TitanQ API documentation <https://docs.titanq.infinityq.io/>`_

        Parameters
        ----------
        beta
            Scales the problem by this factor (inverse of temperature). Beta values can then be
            adjusted to see if a better objective function value can be obtained.
            A lower beta allows for easier escape from local minima, while a higher beta
            is more likely to respect penalties and constraints.

            `Beta values tuning guide <https://docs.titanq.infinityq.io/user-guide/parameter_tuning_guide#beta>`_

            Range: List of [0, 20000]

            Recommended values: List of [0.004...2]

            NOTE: Beta values should be provided in descending order

            >>> import numpy as np
            >>> num_chains = 8
            >>> beta = (1/(np.linspace(2, 50, num_chains, dtype=np.float32))).tolist()

        coupling_mult
            Strength of parameter that keeps multiple logical copies of variables to have the
            same ground state solution. Heuristic to be tuned for a particular problem. Small values
            of this parameter will lead to an incorrect solution while large values will take a long
            time to converge to the correct solution.

            `coupling_mult tuning guide <https://docs.titanq.infinityq.io/user-guide/parameter_tuning_guide#coupling_mult>`_

            Range: [0, 100]

            Recommended values: [0.05...1.0]

        timeout_in_secs
            Maximum time (in second) the computation can take.

            `timeout_in_secs tuning guide <https://docs.titanq.infinityq.io/user-guide/parameter_tuning_guide#timeout_in_secs>`_

            Range: [0.1, 600]

            NOTE: Currently there is no other stop criteria. All computations will run up to the
            timeout value.

        num_chains
            Number of parallel chains running computation. Only the best result of all
            chains is returned.

            `num_chains tuning guide <https://docs.titanq.infinityq.io/user-guide/parameter_tuning_guide#num_chains--num_engines>`_

            Recommended values: [8, 16, 32]

            NOTE: ``num_chains`` * ``num_engines`` cannot exceed 512

        num_engines
            Number of independent batches of chains to run the computation. The best result
            of the batch of chains in each engine is returned.

            `num_engines tuning guide <https://docs.titanq.infinityq.io/user-guide/parameter_tuning_guide#num_chains--num_engines>`_

            Range: [1, 512]

            NOTE: ``num_chains`` * ``num_engines`` cannot exceed 512

        penalty_scaling
            Scaling factor for constraint penalty violations. Increasing this value results
            in stronger constraint enforcement at the cost of increasing the odds of becoming
            trapped in local minima.

            Range: penalty_scaling > 0

            NOTE: If None, a value will be inferred from the objective function of the problem.

        precision
            Some problems may need a higher precision implementation to converge properly such
            as when problem weights exhibit a high dynamic range. This flag allows this higher
            precision (but slightly slower speed) implementation to be used when ``Precision.HIGH``
            is set. Setting ``Precision.STANDARD`` uses a medium precision perfect for general use
            and offering the best speed/efficiency. The default setting ``Precision.AUTO`` will
            inspect the problem passed in and determine which of ``Precision.AUTO`` or
            ``Precision.STANDARD`` precision to use based on internal metrics.

        Returns
        -------
        OptimizeResponse
            Optimized response data object

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        MissingObjectiveError
            If no objective matrices have been added to the model.

        Examples
        --------
        basic solve
            >>> response = model.optimize(timeout_in_secs=60)
        multiple engine
            >>> response = model.optimize(timeout_in_secs=60, num_engines=2)
        custom values
            >>> response = model.optimize(beta=[0.1], coupling_mult=0.75, num_chains=8)
        print values
            >>> print("-" * 15, "+", "-" * 26, sep="")
            >>> print("Ising energy   | Result vector")
            >>> print("-" * 15, "+", "-" * 26, sep="")
            >>> for ising_energy, result_vector in response.result_items():
            >>>     print(f"{ising_energy: <14f} | {result_vector}")
        """
        if self._variables.n_variables() == 0:
            raise MissingVariableError("Cannot optimize before adding a variable to the model.")

        if self._objective is None:
            raise MissingObjectiveError("Cannot optimize before adding an objective to the model.")

        result, metrics = self._solve(beta, coupling_mult, timeout_in_secs, num_chains, num_engines, penalty_scaling, precision)

        return OptimizeResponse(self._variables, result, metrics)

    def _solve(
        self,
        beta: List[float],
        coupling_mult: float,
        timeout_in_secs: float,
        num_chains: int,
        num_engines: int,
        penalty_scaling: Optional[float],
        precision: Precision,
    ) -> Tuple[np.ndarray, Dict[str, Any]]:
        """
        issue a solve request and wait for it to complete.

        Parameters
        ----------
        beta
            beta hyper parameter used by the backend solver.
        coupling_mult
            coupling_mult hyper parameter used by the backend solver.
        timeout_in_secs
            Maximum time (in seconds) the backend solver can take to resolve this problem.
        num_chains
            num_chains hyper parameter used by the backend solver.
        num_engines
            num_engines parameter used by the backend solver.

        Returns
        -------
        The result numpy array and the metric json object.
        """
        with self._storage_client.temp_files_manager(
            self._objective.bias(),
            self._objective.weights(),
            self._constraints.bounds() if self._constraints else None,
            self._constraints.weights() if self._constraints else None,
            # variables bounds is only sent if variable is set to binary
            self._variables.variable_bounds(),
        ) as temp_files:

            # api request
            request = api_model.SolveRequest(
                input=temp_files.input(),
                output=temp_files.output(),
                parameters=api_model.Parameters(
                    beta=beta,
                    coupling_mult=coupling_mult,
                    num_chains=num_chains,
                    num_engines=num_engines,
                    penalty_scaling=penalty_scaling,
                    timeout_in_secs=timeout_in_secs,
                    variable_types=self._variables.get_variable_types_str(),
                    precision=str(precision),
                )
            )

            # adding manifest flags to the request
            request.input.manifest = api_model.Manifest(
                has_set_partitioning_constraint=self._manifest.has_set_partitioning_constraint(),
                has_cardinality_constraint=self._manifest.has_cardinality_constraint(),
                has_equality_constraint=self._manifest.has_equality_constraint(),
                has_inequality_constraint=self._manifest.has_inequality_constraint()
            )

            solve_response = self._titanq_client.solve(request)

            # wait for result to be uploaded by the solver and download it
            archive_file_content = temp_files.download_result()
            with zipfile.ZipFile(io.BytesIO(archive_file_content), 'r') as zip_file:
                try:
                    metrics_content = zip_file.read("metrics.json")
                    result_content = zip_file.read("result.npy")
                except KeyError as ex:
                    try:
                        error_content = zip_file.read("error.json")
                        raise OptimizeError(json.loads(error_content)["error"]) from ex
                    except KeyError as e:
                        raise OptimizeError(
                            "Unexpected error in the solver, please contact InfinityQ support for more help" \
                            f" and provide the following computation id {solve_response.computation_id}") from e

        log.debug("Optimization completed")
        return np.load(io.BytesIO(result_content)), json.loads(metrics_content)



# Copyright (c) 2024, InfinityQ Technology, Inc.

from typing import Optional, Tuple, Union

import numpy as np

from .variable_list import VariableVectorList
from .variable import Expression, SubVariable, Term


def process_expression(
    expr: Union[SubVariable, Term, Expression],
    variables: VariableVectorList,
) -> Tuple[Optional[np.ndarray], np.ndarray, float]:
    """
    Processes an expression to extract the bias vector, weights matrix and constant term.

    This function processes an expression and a list of variable vectors to extract
    the bias vector and weights matrix. The bias vector and weights matrix are initialized
    based on the number of variables and then populated based on the terms in the expression.

    Parameters
    ----------
    expr
        The expression from which to extract the bias vector, weights matrix and constant.
    variables
        A list of variable vectors used in the expression.
        Each variable vector's size contributes to the overall size of the bias vector and weights matrix.

    Returns
    -------
    weights matrix if any, bias vector, and constant

    Raises
    ------
    TypeError
        If the expression is of an unsupported type.
    """
    if not isinstance(expr, (SubVariable, Term, Expression)):
        raise TypeError("TitanQ supports expressions that input variable values and output a singular objective")

    total_size = variables.total_variable_size()
    bias = np.zeros(total_size, dtype=np.float32)
    weights = np.zeros((total_size, total_size), dtype=np.float32)

    if isinstance(expr, SubVariable):
        expr = Term(expr, None, 1)
    if isinstance(expr, Term):
        expr = Expression([expr])

    indexed_variables = variables.index_variables()
    constant = 0
    for term in expr.terms():
        if isinstance(term, Term):
            index_v1 = indexed_variables[term.v1().parent().name()] + term.v1().index()
            if term.v2() is None:# linear
                bias[index_v1] += term.coeff()
            else: # quadratic
                index_v2 = indexed_variables[term.v2().parent().name()] + term.v2().index()
                weights[index_v1, index_v2 ] += term.coeff()
                if index_v1 != index_v2:
                    weights[index_v2 , index_v1] += term.coeff()
        else:
            constant -= term

    return weights if np.any(weights) else None, bias, constant